#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <pthread.h>


#define NOW_PER_SEC 1000
#define NBR_SEC       10 // run 10 sec each test

long long int now()
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
//    fprintf(stderr, "now=%lld %lld %lld %lld %lld\n", tv.tv_sec*1000LL+tv.tv_usec/1000LL, tv.tv_sec*1000LL, tv.tv_usec/1000LL, (long long int)tv.tv_sec, (long long int)tv.tv_usec);
    return tv.tv_sec*1000LL+tv.tv_usec/1000LL;
}

pthread_mutex_t mutex=PTHREAD_MUTEX_INITIALIZER;
pthread_spinlock_t spinlock;

static void test_spinlock(int n)
{
    while (n>0)
    {
        pthread_spin_lock(&spinlock);
        pthread_spin_unlock(&spinlock);
        n--;
        pthread_spin_lock(&spinlock);
        pthread_spin_unlock(&spinlock);
        n--;
        pthread_spin_lock(&spinlock);
        pthread_spin_unlock(&spinlock);
        n--;
        pthread_spin_lock(&spinlock);
        pthread_spin_unlock(&spinlock);
        n--;
        pthread_spin_lock(&spinlock);
        pthread_spin_unlock(&spinlock);
        n--;
    }
}

static void test_mutex(int n)
{
    while (n>0)
    {
        pthread_mutex_lock(&mutex);
        pthread_mutex_unlock(&mutex);
        n--;
        pthread_mutex_lock(&mutex);
        pthread_mutex_unlock(&mutex);
        n--;
        pthread_mutex_lock(&mutex);
        pthread_mutex_unlock(&mutex);
        n--;
        pthread_mutex_lock(&mutex);
        pthread_mutex_unlock(&mutex);
        n--;
        pthread_mutex_lock(&mutex);
        pthread_mutex_unlock(&mutex);
        n--;
    }
}

static void test_empty(int n)
{
    while (n>0)
    {
        n--;
        n--;
        n--;
        n--;
        n--;
    }
}

void test(char *name, void (*func)(int n), int n)
{
    int i;

    long long int start, end, stop;
    start=now();
    end=start;
    stop=start+NBR_SEC*NOW_PER_SEC;

//    fprintf(stderr, "%lld %lld %lld\n", start, end, stop);
    i=0;
    while (end<stop)
    {
//        fprintf(stderr, "%lld %lld %lld\n", start, end, stop);
        func(n);
        i++;
        end=now();
    }

    fprintf(stderr, "%s time: %15.1f lock/s\n", name, (float)n*i*NOW_PER_SEC/(end-start));
}


int main(int argc, char *argv[])
{
    int i;
    int n=1000;
    pthread_spin_init(&spinlock, 0);
    test("spinlock   ", test_spinlock, n);
    test("mutex      ", test_mutex, n);
    test("empty      ", test_empty, n);
}
